from __future__ import print_function

#plotting 
import numpy as np
from matplotlib import pyplot as plt

def plot_lifetimes(lifetimes, censorship = None, birthtimes=None, order=False):
    """

    lifetimes: an (n,) numpy array of lifetimes. 
    censorship: an (n,) numpy array of booleans: True if event observed, else False. 
    birthtimes: an (n,) numpy array offsetting the births away from t=0. 


    Creates a lifetime plot, see 
    examples:

    """
    N = lifetimes.shape[0]
    if N>100:
      print("warning: you may want to subsample to less than 100 individuals.")

    if censorship is None:
      censorship = np.ones(N, dtype=bool)

    if birthtimes is None:
      birthtimes = np.zeros(N)

    if order:
       """order by length of lifetimes; probably not very informative."""
       ix = np.argsort(lifetimes,0)
       lifetimes = lifetimes[ix,0]
       censorship = censorship[ix,0]
       birthtimes = birthtimes[ix]

    for i in range(N):
           c = "#A60628" if censorship[i] else "#348ABD"
           plt.hlines( N-1-i, birthtimes[i] , birthtimes[i] + lifetimes[i], color = c, lw=3)
           m = "|" if not censorship[i] else 'o'
           plt.scatter( (birthtimes[i]) + lifetimes[i] ,  N-1-i, color = c, s=30, marker = m)

    plt.ylim(-0.5, N)
    plt.show()
    return


def shaded_plot(x, y, y_upper, y_lower, **kwargs):
    ax = kwargs.pop('ax', plt.gca())
    base_line, = ax.plot(x, y, **kwargs)
    ax.fill_between(x, y_lower, y_upper, facecolor=base_line.get_color(), alpha=0.25)
    return

def plot_regressions(self):
    def plot(ix=None, iloc=None, columns = [], ci_legend=False, ci_force_lines=False, **kwargs):
        """"
        A wrapper around plotting. Matplotlib plot arguments can be passed in, plus:

          flat: a design style with stepped lines and no shading. Similar to R.
          ci_force_lines: force the confidence intervals to be line plots (versus default shaded areas).
          ci_legend: if ci_force_lines, boolean flag to add the line's label to the legend.
          ix: specify a time-based subsection of the curves to plot, ex:
                   .plot(ix=slice(0.,10.)) will plot the time values between t=0. and t=10.   
          iloc: specify a location-based subsection of the curves to plot, ex:
                   .plot(iloc=slice(0,10)) will plot the first 10 time points. 

        """
        assert (ix == None or iloc == None), 'Cannot set both ix and iloc in call to .plot'

        get_method = "ix" if ix != None else "iloc"
        if iloc == ix == None:
            user_submitted_ix = slice(0,None)
        else:
            user_submitted_ix = ix if ix != None else iloc
        get_loc = lambda df: getattr(df, get_method)[user_submitted_ix]

        if len(columns)==0:
          columns = self.cumulative_hazards_.columns

        if "ax" not in kwargs:
            kwargs["ax"] = plt.figure().add_subplot(111)
            
        x = get_loc(self.cumulative_hazards_).index.values.astype(float)
        for column in columns:
            y = get_loc(self.cumulative_hazards_[column]).values
            y_upper = get_loc(self.confidence_intervals_[column].ix['upper']).values
            y_lower = get_loc(self.confidence_intervals_[column].ix['lower']).values
            shaded_plot(x, y, y_upper, y_lower, ax=kwargs["ax"] , label=column)

        return kwargs["ax"] 
    return plot


def plot_dataframes(self, estimate):
    def plot( ix=None, iloc=None, flat=False, show_censors = False, 
               ci_legend=False, ci_force_lines=False, ci_alpha=0.3, ci_show=True,
               bandwidth=None, **kwargs):
        """"
        A wrapper around plotting. Matplotlib plot arguments can be passed in, plus

          flat: an opiniated design style with stepped lines and no shading. Similar to R's plotting. Default: False
          show_censors: place markers at censorship events. Default: False
          ci_alpha: the transparency level of the confidence interval. Default: 0.3
          ci_force_lines: force the confidence intervals to be line plots (versus default shaded areas). Default: False
          ci_show=True: show confidence intervals. Default: True
          ci_legend: if ci_force_lines is True, this is a boolean flag to add the lines' labels to the legend. Default: False
          ix: specify a time-based subsection of the curves to plot, ex:
                   .plot(ix=slice(0.,10.)) will plot the time values between t=0. and t=10.   
          iloc: specify a location-based subsection of the curves to plot, ex:
                   .plot(iloc=slice(0,10)) will plot the first 10 time points. 
          bandwidth: specify the bandwidth of the kernel smoother for the smoothed-hazard rate. Only used 
              when called 'plot_hazard'.

        """
        assert (ix == None or iloc == None), 'Cannot set both ix and iloc in call to .plot().'

        if "ax" not in kwargs:
            kwargs["ax"] = plt.figure().add_subplot(111)
        kwargs['color'] = next( s for s in [kwargs.get('c'), kwargs.get('color'), next(kwargs["ax"]._get_lines.color_cycle)] if s )

        if flat:
            ci_force_lines=True
            kwargs["drawstyle"] = "steps-pre"
            show_censors = True

        if estimate=="hazard_":
            assert bandwidth != None, 'Must specify a bandwidth parameter in the call to plot_hazard.'
            estimate_ = self.smoothed_hazard_(bandwidth)
            confidence_interval_ = self.smoothed_hazard_confidence_intervals_(bandwidth, hazard_=estimate_.values[:,0])
        else:
            confidence_interval_ = getattr(self, 'confidence_interval_')
            estimate_ = getattr(self, estimate)
                  
        get_method = "ix" if ix != None else "iloc"
        if iloc == ix == None:
            user_submitted_ix = slice(0,None)
        else:
            user_submitted_ix = ix if ix != None else iloc

        get_loc = lambda df: getattr(df, get_method)[user_submitted_ix]

        if show_censors:
            v =  get_loc(estimate_.ix[~(self.event_times['censored'] > 0)])
            kwargs['ax'].plot( v.index.values.astype(float), v.values[:,0], marker="o", color=kwargs['color'], linestyle='None', ms=6, mew=0 )

        get_loc(estimate_).plot(**kwargs)

        if ci_show:
          if ci_force_lines:
              get_loc(confidence_interval_).plot(linestyle="--", linewidth=1, 
                                                 c=kwargs['color'], legend=ci_legend, 
                                                 drawstyle=kwargs.get('drawstyle','default'), ax=kwargs['ax'])
          else:
              x = get_loc(confidence_interval_).index.values.astype(float)
              lower = get_loc(confidence_interval_.filter(like='lower')).values[:,0]
              upper = get_loc(confidence_interval_.filter(like='upper')).values[:,0]
              kwargs['ax'].fill_between(x, lower, y2=upper, alpha=ci_alpha, color=kwargs['color'])
        return kwargs['ax']
    return plot




from __future__ import print_function

from datetime import datetime

import numpy as np
import pandas as pd
from pandas import to_datetime

def group_event_series( groups, durations, censorship, limit=-1):
    """
    Joins multiple event series together into dataframes. A generalization of 
    `survival_table_from_events` to data with groups.

    Parameters:
        groups: a (n,) array of individuals' group ids.
        durations: a (n,) array of durations of each individual
        censorship: a (n,) array of censorship, 1 if observed, 0 else.  

    Output:
        - np.array of unique groups
        - dataframe of removal count data at event_times for each group, column names are 'removed:<group name>'
        - dataframe of observed count data at event_times for each group, column names are 'observed:<group name>'
        - dataframe of censored count data at event_times for each group, column names are 'censored:<group name>'

    Example:
        #input
        group_event_series(waltonG, waltonT, np.ones_like(waltonT)) #data available in test_suite.py

        #output 
        [ 
            array(['control', 'miR-137'], dtype=object), 

                      removed:control  removed:miR-137
            event_at
            6                       0                1
            7                       2                0
            9                       0                3
            13                      0                3
            15                      0                2
            ...,
                      observed:control  observed:miR-137
            event_at
            6                        0                 1
            7                        2                 0
            9                        0                 3
            13                       0                 3
            15                       0                 2
            ...,
                      censored:control  censored:miR-137
            event_at
            6                        0                 0
            7                        0                 0
            9                        0                 0
            ...,
        ]

    """
    unique_groups = np.unique(groups)

    #set first group
    g = unique_groups[0]
    ix = groups == g
    T = durations[ix]
    C = censorship[ix]

    g_name = str(g)
    data = survival_table_from_events(T,C,columns=['removed:'+g_name, "observed:"+g_name, 'censored:'+g_name])
    for g in unique_groups[1:]:
        ix = groups == g
        T = durations[ix]
        C = censorship[ix]
        g_name = str(g)
        data = data.join(survival_table_from_events(T,C,columns=['removed:'+g_name, "observed:"+g_name, 'censored:'+g_name]), how='outer' )
    data = data.fillna(0)
    #hmmm pandas...its too bad I can't do data.ix[:limit] and leave out the if. 
    if int(limit) != -1:
        data = data.ix[:limit]
    return unique_groups, data.filter(like='removed:'), data.filter(like='observed:'), data.filter(like='censored:')


def survival_table_from_events(event_times, censorship, columns=["removed", "observed", "censored"], weights=None):
    """
    Parameters:
        event_times: (n,1) array of event times 
        censorship: if not None, (n,1) boolean array, 1 if observed event, 0 is censored
        columns: a 3-length array to call the, in order, removed individuals, observed deaths
          and censorships.

    Returns:
        Pandas DataFrame with index as the unique times in event_times. The columns named 
        'removed' refers to the number of individuals who were removed from the population 
        by the end of the period. The column 'observed' refers to the number of removed 
        individuals who were observed to have died (i.e. not censored.) The column
        'censored' is defined as 'removed' - 'observed' (the number of individuals who
         left the population due to censorship)
    
    Example:
        #input
        survival_table_from_events( waltonT, np.ones_like(waltonT)) #available in test suite

        #output

                  removed  observed  censored
        event_at
        6               1         1         0
        7               2         2         0
        9               3         3         0
        13              3         3         0
        15              2         2         0

    """
    event_times = np.array( event_times )
    df = pd.DataFrame( event_times.astype(float), columns=["event_at"] )
    df[columns[0]] = 1 if weights == None else weights
    df[columns[1]] = censorship
    event_times = df.groupby("event_at").sum().sort_index()
    event_times[columns[2]] = event_times[columns[0]] - event_times[columns[1]]
    return event_times

def datetimes_to_durations( start_times, end_times, fill_date=datetime.today(),  freq='D', dayfirst=False, na_values=None ):
    """
    This is a very flexible function for transforming arrays of start_times and end_times 
    to the proper format for lifelines: duration and censorship arrays.

    Parameters:
        start_times: an array, series or dataframe of start times. These can be strings, or datetimes.
        end_times: an array, series or dataframe of end times. These can be strings, or datetimes.
                   These values can be None, or an empty string, which corresponds to censorship.
        fill_date: the date to use if end_times is a None or empty string. This corresponds to last date 
                  of observation. Anything above this date is also censored. Default: datetime.today()
        freq: the units of time to use.  See pandas 'freq'. Default 'D' for days.
        day_first: convert assuming European-style dates, i.e. day/month/year. 
        na_values : Additional string to recognize as NA/NaN. Ex: ['']

    Returns:
        T: a array of floats representing the durations with time units given by freq.
        C: a boolean array of censorship: 1 if death observed, 0 else. 

    """
    freq_string = 'timedelta64[%s]'%freq
    start_times = pd.Series(start_times).copy()
    end_times = pd.Series(end_times).copy()
    start_times_ = to_datetime(start_times, dayfirst=dayfirst)

    C = ~( pd.isnull(end_times).values + (end_times=="") + (end_times==na_values) )
    end_times[~C] = fill_date 
    """
    c =  (to_datetime(end_times, dayfirst=dayfirst, coerce=True) > fill_date)
    end_times[c] = fill_date
    C += c
    """
    end_times_ = to_datetime(end_times, dayfirst=dayfirst, coerce=True)

    T = (end_times_ - start_times_).map(lambda x: x.astype(freq_string).astype(float) )
    if (T < 0).sum():
        print("Warning: some values of start_times are before end_times")
    return T.values, C.values

def inv_normal_cdf(p):
    if p < 0.5:
      return -AandS_approximation(p)
    else:
      return AandS_approximation(1-p)

def AandS_approximation(p):
    #Formula 26.2.23 from A&S and help from John Cook ;)
    # http://www.johndcook.com/normal_cdf_inverse.html
    c_0 = 2.515517
    c_1 = 0.802853
    c_2 = 0.010328

    d_1 = 1.432788
    d_2 = 0.189269
    d_3 = 0.001308

    t = np.sqrt(-2*np.log(p))

    return t - (c_0+c_1*t+c_2*t**2)/(1+d_1*t+d_2*t*t+d_3*t**3)


def median_loss(T, T_pred):
    return np.abs(T-T_pred).mean()

def cross_validation(fitter, T, X, censorship=None, k=5, loss_function="median"):
    """
    This implements the censor-sensative loss function as given in 
    'Prediction Performance of Survival Models', Yan Yuan, 2008
    """
    pass

def yaun_loss(fitter, T_true, T_pred, X_train, T_train, censorship=None):
    "[WIP]"
    if censorship==None:
        t = T_pred.shape[0]
        sC = np.ones((t,1))
        censorship = np.ones((t,1))
    else:
        #We are estimating S_c(*|z),is the probability that the 
        # jth individual survived to time yj without being censored
        fitter.fit(T_train,X_train, censorship=1-censorship, timeline=T_true) 
        sC = fitter.predict_survival_function(X_train).ix[T_true[:,0]].values.diagonal()[:,None] #dirty way to get this

    return (censorship*(T_pred-T_true)**2/sC).mean()

def quadrature(fx,x):
    t = x.shape[0]
    return (0.5*fx[:,0] + fx[:,1:-1].sum(1) + 0.5*fx[:,-1])*(x[-1] - x[0])*1.0/t

def basis(n,i):
    x = np.zeros((n,1))
    x[i] = 1
    return x

def epanechnikov_kernel(t, T, bandwidth=1.):
    M = 0.75*(1-(t-T)/bandwidth)**2
    M[ abs((t-T)) >= bandwidth] = 0
    return M



import numpy as np
from numpy.linalg import LinAlgError, inv, pinv
from numpy import dot
import pandas as pd

from lifelines.plotting import plot_dataframes, plot_regressions
from lifelines.utils import survival_table_from_events, basis, inv_normal_cdf, quadrature, epanechnikov_kernel

class NelsonAalenFitter(object):
    """
    Class for fitting the Nelson-Aalen estimate for the cumulative hazard.

    NelsonAalenFitter( alpha=0.95, nelson_aalen_smoothing=True)

    alpha: The alpha value associated with the confidence intervals.
    nelson_aalen_smoothing: If the event times are naturally discrete (like discrete years, minutes, etc.)
      then it is advisable to turn this parameter to False. See [1], pg.84.

    """
    def __init__(self, alpha=0.95, nelson_aalen_smoothing=True):
        self.alpha = alpha
        self.nelson_aalen_smoothing = nelson_aalen_smoothing
        if nelson_aalen_smoothing:
          self._variance_f = self._variance_f_smooth
          self._additive_f = self._additive_f_smooth
        else:
          self._variance_f = self._variance_f_discrete
          self._additive_f = self._additive_f_discrete

    def fit(self, event_times,censorship=None, timeline=None, columns=['NA-estimate'], alpha=None, insert_0=True):
        """
        Parameters:
          event_times: an (n,1) array of times that the death event occured at
          timeline: return the best estimate at the values in timelines (postively increasing)
          columns: a length 1 array to name the column of the estimate.
          alpha: the alpha value in the confidence intervals. Overrides the initializing
             alpha for this call to fit only.
          insert_0: add a leading 0 (if not present) in the timeline.


        Returns:
          DataFrame with index either event_times or timelines (if not None), with
          values as the NelsonAalen estimate
        """

        if censorship is None:
           self.censorship = np.ones(len(event_times), dtype=bool)
        else:
           self.censorship = np.array(censorship).copy().astype(bool)
           
        self.event_times = survival_table_from_events(event_times, self.censorship)

        if alpha is None:
            alpha = self.alpha

        if timeline is None:
           self.timeline = self.event_times.index.values.copy().astype(float)
           if insert_0 and self.timeline[0]>0:
               self.timeline = np.insert(self.timeline,0,0.)
        else:
           self.timeline = timeline.astype(float)

        cumulative_hazard_, cumulative_sq_ = _additive_estimate(self.event_times, self.timeline,
                                                                     self._additive_f, self._variance_f )
        self.cumulative_hazard_ = pd.DataFrame(cumulative_hazard_, columns=columns)
        self.confidence_interval_ = self._bounds(cumulative_sq_[:,None],alpha)
        self.plot = plot_dataframes(self, "cumulative_hazard_")
        self.plot_cumulative_hazard = self.plot
        self.plot_hazard = plot_dataframes(self, 'hazard_')
        self._cumulative_sq = cumulative_sq_

        return self

    def _bounds(self, cumulative_sq_, alpha):
        alpha2 = inv_normal_cdf(1 - (1-alpha)/2)
        df = pd.DataFrame( index=self.timeline)
        name = self.cumulative_hazard_.columns[0]
        df["%s_upper_%.2f"%(name,self.alpha)] = self.cumulative_hazard_.values*np.exp(alpha2*np.sqrt(cumulative_sq_)/self.cumulative_hazard_.values )
        df["%s_lower_%.2f"%(name,self.alpha)] = self.cumulative_hazard_.values*np.exp(-alpha2*np.sqrt(cumulative_sq_)/self.cumulative_hazard_.values )
        return df

    def _variance_f_smooth(self, population, deaths):
        df = pd.DataFrame( {'N':population, 'd':deaths})
        return df.apply( lambda N_d: np.sum([1./(N_d[0]-i)**2 for i in range(int(N_d[1]))]), axis=1 )

    def _variance_f_discrete(self, population, deaths):
        return 1.*(population-deaths)*deaths/population**3

    def _additive_f_smooth(self, population, deaths):
        df = pd.DataFrame( {'N':population, 'd':deaths})
        return df.apply( lambda N_d: np.sum([1./(N_d[0]-i) for i in range(int(N_d[1]))]), axis=1 )

    def _additive_f_discrete(self, population, deaths):
       return (1.*deaths/population).replace([np.inf],0)

    def smoothed_hazard_(self, bandwidth):
        """
        bandwidth: the bandwith used in the Epanechnikov kernel.

        """
        timeline = self.timeline
        cumulative_hazard_name =  self.cumulative_hazard_.columns[0]
        hazard_name = "smoothed-" + cumulative_hazard_name
        hazard_ = self.cumulative_hazard_.diff().fillna(self.cumulative_hazard_.iloc[0] )
        C = (hazard_[cumulative_hazard_name] != 0.0).values
        return pd.DataFrame( 1./(2*bandwidth)*np.dot(epanechnikov_kernel(timeline[:,None], timeline[C][None,:],bandwidth), hazard_.values[C,:]),
                columns=[hazard_name], index=timeline)


    def smoothed_hazard_confidence_intervals_(self, bandwidth, hazard_=None):
        """
        bandwidth: the bandwith to use in the Epanechnikov kernel.
        hazard_: a computed (n,) numpy array of estimated hazard rates. If none, uses naf.smoothed_hazard_
        """
        if hazard_==None:
          hazard_ = self.smoothed_hazard_(bandwidth).values[:,0]

        timeline = self.timeline
        alpha2 = inv_normal_cdf(1 - (1-self.alpha)/2)
        name = "smoothed-" + self.cumulative_hazard_.columns[0]
        self._cumulative_sq.iloc[0] = 0
        var_hazard_ = self._cumulative_sq.diff().fillna( self._cumulative_sq.iloc[0] )
        C = (var_hazard_.values != 0.0) #only consider the points with jumps
        std_hazard_ = np.sqrt(1./(2*bandwidth**2)*np.dot(epanechnikov_kernel(timeline[:,None], timeline[C][None,:],bandwidth)**2, var_hazard_.values[C]))
        values = {
              "%s_upper_%.2f"%(name,self.alpha):hazard_*np.exp(alpha2*std_hazard_/hazard_),
              "%s_lower_%.2f"%(name,self.alpha):hazard_*np.exp(-alpha2*std_hazard_/hazard_)
              }
        return pd.DataFrame(values, index=timeline)

    def __repr__(self):
      try:
        s="""<lifelines.NelsonAalenFitter: fitted with %d observations, %d censored. \n"""%(self.censorship.shape[0], (~self.censorship).sum()) 
        s += self.cumulative_hazard_.head(4).to_string(col_space=15) + "\n... >"
      except AttributeError as e:
        s= """<lifelines.NelsonAalenFitter>"""
      return s

class KaplanMeierFitter(object):

  def __init__(self, alpha=0.95):
       self.alpha = alpha

  def fit(self, event_times, censorship=None, timeline=None, columns=['KM-estimate'], alpha=None, insert_0=True):
       """
       Parameters:
         event_times: an (n,1) array of times that the death event occured at
         timeline: return the best estimate at the values in timelines (postively increasing)
         censorship: an (n,1) array of booleans -- True if the the death was observed, False if the event
            was lost (right-censored). Defaults all True if censorship==None
         columns: a length 1 array to name the column of the estimate.
         alpha: the alpha value in the confidence intervals. Overrides the initializing
            alpha for this call to fit only.
         insert_0: add a leading 0 (if not present) in the timeline.

       Returns:
         DataFrame with index either event_times or timelines (if not None), with
         values under column_name with the KaplanMeier estimate
       """
       #set to all observed if censorship is none
       if censorship is None:
          self.censorship = np.ones(len(event_times), dtype=bool)
       else:
          self.censorship = np.array(censorship).copy().astype(bool)

       if not alpha:
          alpha = self.alpha

       self.event_times = survival_table_from_events(event_times, self.censorship)

       if timeline is None:
          self.timeline = self.event_times.index.values.copy().astype(float)
          if insert_0 and self.timeline[0]>0:
              self.timeline = np.insert(self.timeline,0,0.)
       else:
          self.timeline = timeline.astype(float)

       log_survival_function, cumulative_sq_ = _additive_estimate(self.event_times, self.timeline,
                                                                  self._additive_f, self._additive_var)

       self.survival_function_ = pd.DataFrame(np.exp(log_survival_function), columns=columns)
       #self.median_ = median_survival_times(self.survival_function_)
       self.confidence_interval_ = self._bounds(cumulative_sq_[:,None],alpha)
       self.plot = plot_dataframes(self, "survival_function_")
       self.plot_survival_function = self.plot
       self.median_ = median_survival_times(self.survival_function_)
       return self

  def _bounds(self, cumulative_sq_, alpha):
      # See http://courses.nus.edu.sg/course/stacar/internet/st3242/handouts/notes2.pdfg
      alpha2 = inv_normal_cdf((1.+ alpha)/2.)
      df = pd.DataFrame( index=self.timeline)
      name = self.survival_function_.columns[0]
      v = np.log(self.survival_function_.values)
      df["%s_upper_%.2f"%(name,self.alpha)] =  np.exp(-np.exp(np.log(-v)+alpha2*np.sqrt(cumulative_sq_)/v))
      df["%s_lower_%.2f"%(name,self.alpha)] =  np.exp(-np.exp(np.log(-v)-alpha2*np.sqrt(cumulative_sq_)/v))
      return df

  def _additive_f(self, population, deaths):
     np.seterr(invalid='ignore')
     return (np.log(population-deaths) - np.log(population))

  def _additive_var(self, population, deaths):
      np.seterr(divide='ignore')
      return (1.*deaths/(population*(population-deaths))).replace([np.inf],0)

  def __repr__(self):
    try:
      s="""<lifelines.KaplanMeierFitter: fitted with %d observations, %d censored. \n"""%(self.censorship.shape[0], (~self.censorship).sum()) 
      s += self.survival_function_.head(4).to_string(col_space=15) + "\n... >"
    except AttributeError as e:
      s= """<lifelines.KaplanMeierFitter>"""
    return s


def _additive_estimate(events, timeline, _additive_f, _additive_var):
    """
    Called to compute the Kaplan Meier and Nelson-Aalen estimates.

    """
    N = events["removed"].sum()

    deaths = events['observed']
    population = N - events['removed'].cumsum().shift(1).fillna(0)
    estimate_ = np.cumsum(_additive_f(population,deaths))
    var_ = np.cumsum(_additive_var(population, deaths))

    estimate_ = estimate_.reindex(timeline, method='pad').fillna(0)
    var_ = var_.reindex(timeline, method='pad')
    var_.index.name='timeline'
    estimate_.index.name='timeline'
    
    return estimate_, var_


class AalenAdditiveFitter(object):
    """
    This class fits the regression model:

    hazard(t)  = b_0(t) + b_t(t)*x_1 + ... + b_N(t)*x_N

    that is, the hazard rate is a linear function of the covariates. Currently the covariates
    must be time independent.

    Parameters:
      fit_intercept: If False, do not attach an intercept (column of ones) to the covariate matrix. The
        intercept, b_0(t) acts as a baseline hazard.
      alpha: the level in the confidence intervals.
      penalizer: Attach a L2 penalizer to the regression. This improves stability of the estimates
       and controls high correlation between covariates.

    """

    def __init__(self,fit_intercept=True, alpha=0.95, penalizer = 0.0):
        self.fit_intercept = fit_intercept
        self.alpha = alpha
        self.penalizer = penalizer
        assert penalizer >= 0, "penalizer keyword must be >= 0."

    def fit(self, event_times, X, timeline=None, censorship=None, columns=None):
        """currently X is a static (n,d) array

        event_times: (n,1) array of event times
        X: (n,d) the design matrix, either a numpy matrix or DataFrame.
        timeline: (t,1) timepoints in ascending order
        censorship: (n,1) boolean array of censorships: True if observed, False if right-censored.
                    By default, assuming all are observed.

        Fits: self.cumulative_hazards_: a (t,d+1) dataframe of cumulative hazard coefficients
              self.hazards_: a (t,d+1) dataframe of hazard coefficients

        """
        #deal with the covariate matrix. Check if it is a dataframe or numpy array
        n,d = X.shape
        if type(X)==pd.core.frame.DataFrame:
          X_ = X.values.copy()
          if columns is None:
            columns = X.columns
        else:
          X_ = X.copy()

        # append a columns of ones for the baseline hazard
        ix = event_times.argsort(0)[:,0].copy()
        X_ = X_[ix,:].copy() if not self.fit_intercept else np.c_[ X_[ix,:].copy(), np.ones((n,1)) ]
        sorted_event_times = event_times[ix,0].copy()

        #set the column's names of the dataframe.
        if columns is None:
          columns = range(d)
        else:
          columns = [c for c in columns]

        if self.fit_intercept:
          columns += ['baseline']

        #set the censorship events. 1 if the death was observed.
        if censorship is None:
            observed = np.ones(n, dtype=bool)
        else:
            observed = censorship[ix].reshape(n)

        #set the timeline -- this is used as DataFrame index in the results
        if timeline is None:
            timeline = sorted_event_times.copy()

        timeline = np.unique(timeline.astype(float))
        if timeline[0] > 0:
           timeline = np.insert(timeline,0,0.)

        unique_times = np.unique(timeline)
        zeros = np.zeros((timeline.shape[0],d+self.fit_intercept))
        self.cumulative_hazards_ = pd.DataFrame(zeros.copy(), index=unique_times, columns = columns)
        self.hazards_ = pd.DataFrame(np.zeros((event_times.shape[0],d+self.fit_intercept)), index=event_times[:,0], columns = columns)
        self._variance = pd.DataFrame(zeros.copy(), index=unique_times, columns = columns)

        #create the penalizer matrix for L2 regression
        penalizer = self.penalizer*np.eye(d + self.fit_intercept)

        t_0 = sorted_event_times[0]
        cum_v = np.zeros((d+self.fit_intercept,1))
        v = cum_v.copy()
        for i,time in enumerate(sorted_event_times):
            relevant_times = (t_0<timeline)*(timeline<=time)
            if observed[i] == 0:
                X_[i,:] = 0
            try:
                V = dot(inv(dot(X_.T,X_) + penalizer), X_.T)
            except LinAlgError:
                #if penalizer > 0, this should not occur. But sometimes it does...
                V = dot(pinv(dot(X_.T,X_) + penalizer), X_.T)

            v = dot(V, basis(n,i))
            cum_v = cum_v + v
            self.cumulative_hazards_.ix[relevant_times] = self.cumulative_hazards_.ix[relevant_times].values + cum_v.T
            self.hazards_.iloc[i] = self.hazards_.iloc[i].values + v.T
            self._variance.ix[relevant_times] = self._variance.ix[relevant_times].values + dot( V[:,i][:,None], V[:,i][None,:] ).diagonal()
            t_0 = time
            X_[i,:] = 0

        #clean up last iteration
        relevant_times = (timeline>time)
        self.hazards_.iloc[i] = v.T
        try:
            self.cumulative_hazards_.ix[relevant_times] = cum_v.T
            self._variance.ix[relevant_times] = dot( V[:,i][:,None], V[:,i][None,:] ).diagonal()
        except:
            pass
        self.timeline = timeline
        self.X = X
        self.censorship = censorship
        self.event_times = event_times
        self._compute_confidence_intervals()
        self.plot = plot_regressions(self)
        return self

    def smoothed_hazards_(self, bandwith=1):
        """
        Using the epanechnikov kernel to smooth the hazard function, with sigma/bandwith

        """
        C = self.censorship.astype(bool)
        return pd.DataFrame( np.dot(epanechnikov_kernel(self.timeline[:,None], self.timeline[C],bandwith), self.hazards_.values[C,:]),
                columns=self.hazards_.columns, index=self.timeline)

    def _compute_confidence_intervals(self):
        alpha2 = inv_normal_cdf(1 - (1-self.alpha)/2)
        n = self.timeline.shape[0]
        d = self.cumulative_hazards_.shape[1]
        index = [['upper']*n+['lower']*n, np.concatenate( [self.timeline, self.timeline] ) ]
        self.confidence_intervals_ = pd.DataFrame(np.zeros((2*n,d)), index=index, columns=self.cumulative_hazards_.columns)
        self.confidence_intervals_.ix['upper'] = self.cumulative_hazards_.values + alpha2*np.sqrt(self._variance.cumsum().values)
        self.confidence_intervals_.ix['lower'] = self.cumulative_hazards_.values - alpha2*np.sqrt(self._variance.cumsum().values)
        return

    def predict_cumulative_hazard(self, X, columns=None):
        """
        X: a (n,d) covariate matrix

        Returns the hazard rates for the individuals
        """
        n,d = X.shape
        try:
          X_ = X.values.copy()
        except:
          X_ = X.copy()
        X_ = X.copy() if not self.fit_intercept else np.c_[ X.copy(), np.ones((n,1)) ]
        return pd.DataFrame(np.dot(self.cumulative_hazards_, X_.T), index=self.timeline, columns=columns )

    def predict_survival_function(self,X, columns=None):
        """
        X: a (n,d) covariate matrix

        Returns the survival functions for the individuals
        """
        return np.exp(-self.predict_cumulative_hazard(X, columns=columns))

    def predict_median(self,X):
        """
        X: a (n,d) covariate matrix
        Returns the median lifetimes for the individuals
        """
        return median_survival_times(self.predict_survival_function(X))

    def predict_expectation(self,X):
        """
        Compute the expected lifetime, E[T], using covarites X.
        """
        t = self.cumulative_hazards_.index
        return quadrature(self.predict_survival_function(X).values.T, t)

#utils
def qth_survival_times(q, survival_functions):
    """
    This can be done much better.

    Parameters:
      q: a float between 0 and 1.
      survival_functions: a (n,d) dataframe or numpy array.
        If dataframe, will return index values (actual times)
        If numpy array, will return indices.

    Returns:
      v: an array containing the first times the value was crossed.
        np.inf if infinity.
    """
    assert 0.<=q<=1., "q must be between 0. and 1."
    sv_b = (1.0*(survival_functions < q)).cumsum() > 0
    try:
        v = sv_b.idxmax(0)
        v[sv_b.iloc[-1,:]==0] = -1
    except:
        v = sv_b.argmax(0)
        v[sv_b[-1,:]==0] = np.inf
    return v

def median_survival_times(survival_functions):
    return qth_survival_times(0.5, survival_functions)

def asymmetric_epanechnikov_kernel(q, x):
    return (64*(2 - 4*q + 6*q*q - 3*q**3) + 240*(1-q)**2*x)/((1+q)**4*(19 - 18*q + 3*q**2))

"""
References:
[1] Aalen, O., Borgan, O., Gjessing, H., 2008. Survival and Event History Analysis

"""



"""
python -m lifelines.tests.test_suit

"""
from __future__ import print_function

import unittest

import numpy as np
import numpy.testing as npt
from collections import Counter
import matplotlib.pyplot as plt
import pandas as pd

from ..estimation import KaplanMeierFitter, NelsonAalenFitter, AalenAdditiveFitter, median_survival_times
from ..statistics import logrank_test, multivariate_logrank_test, pairwise_logrank_test
from ..generate_datasets import *
from ..plotting import plot_lifetimes
from ..utils import datetimes_to_durations

class MiscTests(unittest.TestCase):

    def test_quadrature_integration_for_identity_fuction(self):
        #integrate x between 0 and 10:
        x = np.linspace(0,10,1000)[:,None]
        answer = 0.5*x**2
        approx_answer = cumulative_quadrature( x.T, x).T
        npt.assert_almost_equal(answer, approx_answer, decimal=5)

    def test_quadrature_integration_for_exp_function(self):
        #integrate exp(x) between 0 and 4:
        x = np.linspace(0,4,1000)[:,None]
        answer = np.exp(x) -1.
        approx_answer = cumulative_quadrature( np.exp(x).T, x).T
        #pdb.set_trace()
        npt.assert_almost_equal(answer, approx_answer, decimal=4)

    def test_aalen_additive_allows_numpy_or_df(self):
        t = np.random.random((10,1))
        dfX = pd.DataFrame(np.random.random((10,3)), columns = ["A", "B", "C"])
        c = dfX.columns
        npX = np.random.random((10,5))
        aaf = AalenAdditiveFitter()
        aaf.fit(t,npX)
        aaf.fit(t, dfX)
        npt.assert_array_equal( c, aaf.cumulative_hazards_.columns[:3] )
    
    def test_datetimes_to_durations_days(self):
        start_date = ['2013-10-10 0:00:00', '2013-10-09', '2012-10-10']
        end_date = ['2013-10-13', '2013-10-10 0:00:00', '2013-10-15']
        T,C = datetimes_to_durations(start_date, end_date)
        npt.assert_almost_equal(T, np.array([3,1,5+365]) )
        npt.assert_almost_equal(C, np.array([1,1,1], dtype=bool) )
        return

    def test_datetimes_to_durations_years(self):
        start_date = ['2013-10-10', '2013-10-09', '2012-10-10']
        end_date = ['2013-10-13', '2013-10-10', '2013-10-15']
        T,C = datetimes_to_durations(start_date, end_date, freq='Y')
        npt.assert_almost_equal(T, np.array([0,0,1]) )
        npt.assert_almost_equal(C, np.array([1,1,1], dtype=bool) )
        return

    def test_datetimes_to_durations_hours(self):
        start_date = ['2013-10-10 17:00:00', '2013-10-09 0:00:00', '2013-10-10 23:00:00']
        end_date = ['2013-10-10 18:00:00', '2013-10-10 0:00:00', '2013-10-11 2:00:00']
        T,C = datetimes_to_durations(start_date, end_date, freq='h')
        npt.assert_almost_equal(T, np.array([1,24,3]) )
        npt.assert_almost_equal(C, np.array([1,1,1], dtype=bool) )
        return
    
    def test_datetimes_to_durations_censor(self):
        start_date = ['2013-10-10', '2013-10-09', '2012-10-10']
        end_date = ['2013-10-13', None, '']
        T,C = datetimes_to_durations(start_date, end_date, freq='Y')
        npt.assert_almost_equal(C, np.array([1,0,0], dtype=bool) )
        return

    def test_datetimes_to_durations_custom_censor(self):
        start_date = ['2013-10-10', '2013-10-09', '2012-10-10']
        end_date = ['2013-10-13', "NaT", '']
        T,C = datetimes_to_durations(start_date, end_date, freq='Y', na_values="NaT")
        npt.assert_almost_equal(C, np.array([1,0,0], dtype=bool) )
        return

    """
    def test_datetimes_to_durations_max_dates(self):
        start_date = ['2013-10-10', '2013-10-09', '2013-10-08']
        end_date =  ['2014-10-10', '2014-10-09', '2014-10-08']
        T,C = datetimes_to_durations(start_date, end_date, fill_date = "2014-10-09" )
        npt.assert_almost_equal(C, np.array([1,0,0], dtype=bool) )
        return
    """
class StatisticalTests(unittest.TestCase):

  def setUp(self):
      self.lifetimes = Counter(LIFETIMES)
      self.km = self.kaplan_meier()
      self.kmc = self.kaplan_meier(censor=True)
      self.na = self.nelson_aalen()
      self.nac = self.nelson_aalen(censor=True)

  def test_kaplan_meier(self):
      kmf = KaplanMeierFitter()
      kmf.fit(LIFETIMES)
      npt.assert_almost_equal(kmf.survival_function_.values, self.km )

  def test_nelson_aalen(self):
      naf = NelsonAalenFitter(nelson_aalen_smoothing=False)
      naf.fit(LIFETIMES)
      npt.assert_almost_equal(naf.cumulative_hazard_.values, self.na )

  def test_censor_nelson_aalen(self):
      naf = NelsonAalenFitter(nelson_aalen_smoothing=False)
      naf.fit(LIFETIMES, censorship=CENSORSHIP)
      npt.assert_almost_equal(naf.cumulative_hazard_.values, self.nac )

  def test_censor_kaplan_meier(self):
      kmf = KaplanMeierFitter()
      kmf.fit(LIFETIMES, censorship = CENSORSHIP)
      npt.assert_almost_equal(kmf.survival_function_.values, self.kmc )

  def test_median(self):
      sv = pd.DataFrame(1 - np.linspace(0,1,1000))
      self.assertTrue( median_survival_times(sv).ix[0] == 500 )

  def test_not_to_break(self):
      try:
        naf = NelsonAalenFitter(nelson_aalen_smoothing=False)
        naf.fit(LIFETIMES)
        naf = NelsonAalenFitter(nelson_aalen_smoothing=True)
        naf.fit(LIFETIMES)
        self.assertTrue(True)
      except Exception as e:
        print(e)
        self.assertTrue(False)

  def test_equal_intensity(self):
      data1 = np.random.exponential(5, size=(200,1))
      data2 = np.random.exponential(5, size=(200,1))
      summary, p_value, result = logrank_test(data1, data2)
      print(summary)
      self.assertTrue(result==None)

  def test_unequal_intensity(self):
      data1 = np.random.exponential(5, size=(200,1))
      data2 = np.random.exponential(1, size=(200,1))
      summary, p_value, result = logrank_test(data1, data2)
      print(summary)
      self.assertTrue(result)

  def test_unequal_intensity_censorship(self):
      data1 = np.random.exponential(5, size=(200,1))
      data2 = np.random.exponential(1, size=(200,1))
      censorA = np.random.binomial(1,0.5, size=(200,1))
      censorB = np.random.binomial(1,0.5, size=(200,1))
      summary, p_value, result = logrank_test(data1, data2, censorship_A = censorA, censorship_B = censorB)
      print(summary)
      self.assertTrue(result)

  def test_integer_times_logrank_test(self):
      data1 = np.random.exponential(5, size=(200,1)).astype(int)
      data2 = np.random.exponential(1, size=(200,1)).astype(int)
      summary, p_value, result = logrank_test(data1, data2)
      print(summary)
      self.assertTrue(result)

  def test_waltons_data(self):
      summary, p_value, result = logrank_test(waltonT1, waltonT2)
      print(summary)
      self.assertTrue(result)

  def test_smoothing_hazard_ties(self):
      T = np.random.binomial(20,0.7, size=300)
      C = np.random.binomial(1,0.8, size=300)
      naf = NelsonAalenFitter()
      naf.fit(T,C)
      naf.smoothed_hazard_(1.)
      self.assertTrue(True)

  def test_smoothing_hazard_nontied(self):
      T = np.random.exponential(20, size=300)**2
      C = np.random.binomial(1,0.8, size=300)
      naf = NelsonAalenFitter()
      naf.fit(T,C)
      naf.smoothed_hazard_(1.)
      naf.fit(T)
      naf.smoothed_hazard_(1.)
      self.assertTrue(True)

  def test_smoothing_hazard_ties_no_censorship(self):
      T = np.random.binomial(20,0.7, size=300)
      naf = NelsonAalenFitter()
      naf.fit(T)
      naf.smoothed_hazard_(1.)
      self.assertTrue(True)

  def test_smoothing_hazard_with_spike_at_time_0(self):
      T = np.random.binomial(20,0.7, size=300)
      T[np.random.binomial(1,0.3, size=300).astype(bool)] = 0
      naf = NelsonAalenFitter()
      naf.fit(T)
      df = naf.smoothed_hazard_(bandwidth=0.1)
      self.assertTrue(df.iloc[0].values[0] > df.iloc[1].values[0] )

  def test_multivariate_unequal_intensities(self):
      T = np.random.exponential(10, size=300)
      g = np.random.binomial(2, 0.5, size=300)
      T[g==1] = np.random.exponential(6, size=(g==1).sum())
      s, _, result = multivariate_logrank_test(T, g)
      print(s)
      self.assertTrue(result==True)

  def test_multivariate_equal_intensities(self):
      T = np.random.exponential(10, size=300)
      g = np.random.binomial(2, 0.5, size=300)
      s, _, result = multivariate_logrank_test(T, g)
      print(s)
      self.assertTrue(result==None)

  def test_pairwise_waltons_data(self):
      _,_,R = pairwise_logrank_test(waltonT, waltonG)
      print(R)
      self.assertTrue(R.values[0,1])

  def test_pairwise_logrank_test(self):
      T = np.random.exponential(10, size=300)
      g = np.random.binomial(2, 0.7, size=300)
      S, P, R = pairwise_logrank_test(T, g, alpha=0.95)
      V = np.array([[np.nan,None,None],[None,np.nan,None],[None,None,np.nan]])
      npt.assert_array_equal(R,V)

  def test_aalen_additive_median_predictions_split_data(self):
      #This tests to make sure that my median predictions statisfy
      # the prediction are greater than the actual 1/2 the time.
      #generate some hazard rates and a survival data set
      n = 2500
      d = 5
      timeline = np.linspace(0,70,5000)
      hz, coef, X = generate_hazard_rates(n,d, timeline)
      T = generate_random_lifetimes(hz, timeline)
      #fit it to Aalen's model
      aaf = AalenAdditiveFitter(penalizer=0., fit_intercept=False)
      aaf.fit(T,X, censorship=None)

      #predictions
      T_pred = aaf.predict_median(X)
      self.assertTrue( abs((T_pred.values > T).mean() - 0.5) < 0.05 )

  def test_aalen_additive_fit_no_censor(self):
      #this is a visual test of the fitting the cumulative
      #hazards.
      n = 2500
      d = 3
      timeline = np.linspace(0,70,5000)
      hz, coef, X = generate_hazard_rates(n, d, timeline)
      cumulative_hazards = cumulative_quadrature( coef.values.T, timeline).T
      T = generate_random_lifetimes(hz, timeline)

      #fit the aaf, no intercept as it is already built into X, X[2] is ones
      aaf = AalenAdditiveFitter(penalizer=0., fit_intercept=False)
      aaf.fit(T,X, censorship=None, columns = coef.columns)

      T_max = aaf.timeline[-10]
      #plot baby
      for i,column in enumerate(coef.columns):
        ax = plt.subplot(d+2,1,i+1)
        ax.plot( timeline[timeline<T_max], cumulative_hazards[timeline<T_max,i])
        aaf.cumulative_hazards_[column].plot(ax=ax)
        ax.legend(loc='lower right')

      ax = plt.subplot(d+2,1,d+2)
      ax.plot( (T > np.arange(T_max)).sum(0), c="k", label="number of observations" )
      ax.set_xlabel("time")
      plt.show()
      return

  def test_aalen_additive_fit_with_censor(self):
      #this is a visual test of the fitting the cumulative
      #hazards.
      n = 2500
      d = 3
      timeline = np.linspace(0,70,5000)
      hz, coef, X = generate_hazard_rates(n, d, timeline)
      cumulative_hazards = cumulative_quadrature( coef.values.T, timeline).T
      T = generate_random_lifetimes(hz, timeline)
      C = np.random.binomial(1, 0.8, size=n)

      #fit the aaf, no intercept as it is already built into X, X[2] is ones
      aaf = AalenAdditiveFitter(penalizer=0., fit_intercept=False)
      aaf.fit(T,X, censorship=C, columns = coef.columns)

      T_max = aaf.timeline[-10]
      #plot baby
      for i,column in enumerate(coef.columns):
          ax = plt.subplot(d+2,1,i+1)
          ax.plot( timeline[timeline<T_max], cumulative_hazards[timeline<T_max,i])
          aaf.plot(ax=ax, columns=[column],  iloc=slice(0,200))
          ax.legend(loc='lower right')

      ax = plt.subplot(d+2,1,d+2)
      ax.plot( (T > np.arange(T_max)).sum(0), c="k", label="number of observations" )
      ax.set_xlabel("time")
      plt.suptitle("aalen fit with 0.8 observations, d = %d"%d)
      plt.show()
      return

  def test_lists_to_KaplanMeierFitter(self):
      T = [2,3,4.,1.,6,5.]
      C = [1,0,0,0,1,1]
      kmf = KaplanMeierFitter()
      with_list = kmf.fit(T,C).survival_function_.values
      with_array = kmf.fit(np.array(T),np.array(C)).survival_function_.values
      npt.assert_array_equal(with_list,with_array)

  def test_lists_to_NelsonAalenFitter(self):
      T = [2,3,4.,1.,6,5.]
      C = [1,0,0,0,1,1]
      naf = NelsonAalenFitter()
      with_list = naf.fit(T,C).cumulative_hazard_.values
      with_array = naf.fit(np.array(T),np.array(C)).cumulative_hazard_.values
      npt.assert_array_equal(with_list,with_array)

  def test_population_equals_death(self):
      T = [1,1,1]
      kmf = KaplanMeierFitter()
      kmf.fit(T)
      return

  def test_pairwise_allows_dataframes(self):
      N = 100
      df = pd.DataFrame(np.empty((N,3)), columns=["T", "C", "group"])
      df["T"] = np.random.exponential(1,size=N)
      df["C"] = np.random.binomial(1,0.6, size=N)
      df["group"] = np.random.binomial(2,0.5, size=N)
      try:
        pairwise_logrank_test(df['T'], df["group"], censorship=df["C"])
        self.assertTrue(True)
      except:
        self.assertTrue(False)

  def kaplan_meier(self, censor=False):
      km = np.zeros((len(self.lifetimes.keys()),1))
      ordered_lifetimes = np.sort( self.lifetimes.keys())
      v = 1.
      n = N*1.0
      for i,t in enumerate(ordered_lifetimes):
        if censor:
           ix = LIFETIMES == t
           c = sum(1-CENSORSHIP[ix])
           if n!=0:
              v *= ( 1-(self.lifetimes.get(t)-c)/n )
           n -= self.lifetimes.get(t)
        else:
           v *= ( 1-self.lifetimes.get(t)/n )
           n -= self.lifetimes.get(t)
        km[i] = v
      if km[0] < 1.:
        km = np.insert(km,0,1.)
      return km.reshape(len(km),1)

  def nelson_aalen(self, censor = False):
      na = np.zeros((len(self.lifetimes.keys()),1))
      ordered_lifetimes = np.sort( self.lifetimes.keys())
      v = 0.
      n = N*1.0
      for i,t in enumerate(ordered_lifetimes):
        if censor:
           ix = LIFETIMES == t
           c = sum(1-CENSORSHIP[ix])
           if n!=0:
              v += ( (self.lifetimes.get(t)-c)/n )
           n -= self.lifetimes.get(t)
        else:
           v += ( self.lifetimes.get(t)/n )
           n -= self.lifetimes.get(t)
        na[i] = v
      if na[0] > 0:
        na = np.insert(na,0,0.)
      return na.reshape(len(na),1)

class PlottingTests(unittest.TestCase):

  def test_aalen_additive_plot(self):
      #this is a visual test of the fitting the cumulative
      #hazards.
      n = 2500
      d = 3
      timeline = np.linspace(0,70,5000)
      hz, coef, X = generate_hazard_rates(n, d, timeline)
      cumulative_hazards = cumulative_quadrature( coef.values.T, timeline).T
      T = generate_random_lifetimes(hz, timeline)
      C = np.random.binomial(1, 0.8, size=n)

      #fit the aaf, no intercept as it is already built into X, X[2] is ones
      aaf = AalenAdditiveFitter(penalizer=0., fit_intercept=False)
      aaf.fit(T,X, censorship=C, columns = coef.columns)
      ax=aaf.plot(iloc=slice(0, aaf.cumulative_hazards_.shape[0]-100))
      ax.set_xlabel("time")
      ax.set_title('.plot() cumulative hazards')
      return

  def test_aalen_additive_smoothed_plot(self):
      #this is a visual test of the fitting the cumulative
      #hazards.
      n = 2500
      d = 3
      timeline = np.linspace(0,150,5000)
      hz, coef, X = generate_hazard_rates(n, d, timeline)
      cumulative_hazards = cumulative_quadrature( coef.values.T, timeline).T
      T = generate_random_lifetimes(hz, timeline) + 0.1*np.random.uniform(size=(n,1))
      C = np.random.binomial(1, 0.8, size=n)

      #fit the aaf, no intercept as it is already built into X, X[2] is ones
      aaf = AalenAdditiveFitter(penalizer=0., fit_intercept=False)
      aaf.fit(T,X, censorship=C, columns = coef.columns)
      ax=aaf.smoothed_hazards_(1).iloc[0:aaf.cumulative_hazards_.shape[0]-500].plot()
      ax.set_xlabel("time")
      ax.set_title('.plot() smoothed hazards')
      return

  def test_kmf_plotting(self):
      data1 = np.random.exponential(10, size=(100))
      data2 = np.random.exponential(2, size=(200,1))
      data3 = np.random.exponential(4, size=(500,1))
      kmf = KaplanMeierFitter()
      kmf.fit(data1, columns = ['test label 1'])
      ax = kmf.plot()
      kmf.fit(data2, columns = ['test label 2'] )
      kmf.plot(ax=ax)
      kmf.fit(data3, columns = ['test label 3'] )
      kmf.plot(ax=ax)
      plt.title("testing kmf")
      return

  def test_naf_plotting(self):
      data1 = np.random.exponential(5, size=(200,1))
      data2 = np.random.exponential(1, size=(500))
      naf = NelsonAalenFitter()
      naf.fit(data1)
      ax = naf.plot(color="r")
      naf.fit(data2)
      naf.plot(ax=ax, c="k")
      plt.title('testing naf + custom colors')
      return

  def test_estimators_accept_lists(self):
      data = [1,2,3,4,5]
      NelsonAalenFitter().fit(data)
      KaplanMeierFitter().fit(data)
      return True

  def test_ix_slicing(self):
      naf = NelsonAalenFitter().fit(waltonT)
      self.assertTrue(naf.cumulative_hazard_.ix[0:10].shape[0]==4)
      return

  def test_iloc_slicing(self):
      naf = NelsonAalenFitter().fit(waltonT)
      self.assertTrue(naf.cumulative_hazard_.iloc[0:10].shape[0]==10)
      self.assertTrue(naf.cumulative_hazard_.iloc[0:-1].shape[0]==32)
      return

  def test_naf_plotting_slice(self):
      data1 = np.random.exponential(5, size=(200,1))
      data2 = np.random.exponential(1, size=(200,1))
      naf = NelsonAalenFitter()
      naf.fit(data1)
      ax = naf.plot(ix=slice(0,None))
      naf.fit(data2)
      naf.plot(ax=ax, ci_force_lines=True, iloc=slice(100,180))
      plt.title('testing slicing')
      return

  def test_plot_lifetimes_calendar(self):
      plt.figure()
      t = np.linspace(0, 20, 1000)
      hz, coef, covrt =generate_hazard_rates(1,5, t)
      N = 20
      current = 10
      birthtimes = current*np.random.uniform(size=(N,))
      T, C = generate_random_lifetimes(hz, t, size=N, censor=current - birthtimes )
      plot_lifetimes(T, censorship=C, birthtimes=birthtimes)

  def test_plot_lifetimes_relative(self):
      plt.figure()
      t = np.linspace(0, 20, 1000)
      hz, coef, covrt =generate_hazard_rates(1,5, t)
      N = 20
      T, C= generate_random_lifetimes(hz, t, size=N, censor=True )
      plot_lifetimes(T, censorship=C)

  def test_naf_plot_cumulative_hazard(self):
      data1 = np.random.exponential(5, size=(200,1))
      naf = NelsonAalenFitter()
      naf.fit(data1)
      ax = naf.plot()
      naf.plot_cumulative_hazard(ax=ax, ci_force_lines=True)
      plt.title("I should have plotted the same thing, but different styles + color!")
      return

  def test_naf_plot_cumulative_hazard_bandwidth_2(self):
      data1 = np.random.exponential(5, size=(2000,1))
      naf = NelsonAalenFitter()
      naf.fit(data1)
      naf.plot_hazard( bandwidth=1., ix=slice(0,7.))
      plt.title('testing smoothing hazard')
      return

  def test_naf_plot_cumulative_hazard_bandwith_1(self):
      data1 = np.random.exponential(5, size=(2000,1))**2
      naf = NelsonAalenFitter()
      naf.fit(data1)
      naf.plot_hazard( bandwidth=5., iloc=slice(0,1700))
      plt.title('testing smoothing hazard')
      return

  def test_show_censor_with_index_0(self):
      T = np.random.binomial(20,0.1, size=100)
      C = np.random.binomial(1, 0.8, size=100)
      kmf = KaplanMeierFitter()
      kmf.fit(T,C).plot(show_censors=True)
      return


#some data

LIFETIMES = np.array([2,4,4,4,5,7,10,11,11,12])
CENSORSHIP = np.array([1,1,0,1,0,1,1,1,1,0])
N = len(LIFETIMES)

waltonT2 = np.array([6., 13., 13., 13., 19., 19., 19., 26., 26., 26., 26., 26., 33., 33., 47., 62., 62., 9., 9., 9., 15., 15., 22., 22., 22., 22., 29., 29., 29., 29., 29., 36., 36., 43.])
waltonT1 = np.array([33., 54., 54., 61., 61., 61., 61., 61., 61., 61., 61., 61., 61., 61., 69., 69., 69., 69., 69., 69., 69., 69., 69., 69., 69., 32., 53., 53., 60., 60., 60., 60., 60., 68., 68., 68., 68., 68., 68., 68., 68., 68., 68., 75., 17., 51., 51., 51., 58., 58., 58., 58., 66., 66., 7., 7., 41., 41., 41., 41., 41., 41., 41., 48., 48., 48., 48., 48., 48., 48., 48., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 56., 63., 63., 63., 63., 63., 63., 63., 63., 63., 69., 69., 38., 38., 45., 45., 45., 45., 45., 45., 45., 45., 45., 45., 53., 53., 53., 53., 53., 60., 60., 60., 60., 60., 60., 60., 60., 60., 60., 60., 66.])

waltonG = np.array(['miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137',
       'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137',
       'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137',
       'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137',
       'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137', 'miR-137',
       'miR-137', 'miR-137', 'miR-137', 'miR-137', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control', 'control', 'control', 'control', 'control', 'control',
       'control'], dtype=object)

waltonT = np.array([  6.,  13.,  13.,  13.,  19.,  19.,  19.,  26.,  26.,  26.,  26.,
        26.,  33.,  33.,  47.,  62.,  62.,   9.,   9.,   9.,  15.,  15.,
        22.,  22.,  22.,  22.,  29.,  29.,  29.,  29.,  29.,  36.,  36.,
        43.,  33.,  54.,  54.,  61.,  61.,  61.,  61.,  61.,  61.,  61.,
        61.,  61.,  61.,  61.,  69.,  69.,  69.,  69.,  69.,  69.,  69.,
        69.,  69.,  69.,  69.,  32.,  53.,  53.,  60.,  60.,  60.,  60.,
        60.,  68.,  68.,  68.,  68.,  68.,  68.,  68.,  68.,  68.,  68.,
        75.,  17.,  51.,  51.,  51.,  58.,  58.,  58.,  58.,  66.,  66.,
         7.,   7.,  41.,  41.,  41.,  41.,  41.,  41.,  41.,  48.,  48.,
        48.,  48.,  48.,  48.,  48.,  48.,  56.,  56.,  56.,  56.,  56.,
        56.,  56.,  56.,  56.,  56.,  56.,  56.,  56.,  56.,  56.,  56.,
        56.,  56.,  63.,  63.,  63.,  63.,  63.,  63.,  63.,  63.,  63.,
        69.,  69.,  38.,  38.,  45.,  45.,  45.,  45.,  45.,  45.,  45.,
        45.,  45.,  45.,  53.,  53.,  53.,  53.,  53.,  60.,  60.,  60.,
        60.,  60.,  60.,  60.,  60.,  60.,  60.,  60.,  66.])


if __name__ == '__main__':
    unittest.main()



#lib to create fake survival datasets
import numpy as np
from numpy import random
import pandas as pd

class coeff_func(object):
  """This is a decorator class used later to construct nice names"""
  def __init__(self, f):
    self.f = f

  def __call__(self, *args, **kwargs):
    def __repr__():
      s = self.f.__doc__.replace("alpha", "%.2f"%kwargs["alpha"]).replace("beta","%.2f"%kwargs["beta"])
      return s
    self.__doc__ = __repr__()
    self.__repr__ = __repr__
    self.__str__ = __repr__
    return self.f(*args, **kwargs)


@coeff_func
def exp_comp_(t,alpha=1,beta =1):
  """beta*(1 - np.exp(-alpha*(t-beta)))"""
  return beta*(1 - np.exp(-alpha*np.maximum(0,t-10*beta)))
@coeff_func
def log_(t,alpha=1, beta=1):
  """beta*np.log(alpha*(t-beta)+1)"""
  return beta*np.log(alpha*np.maximum(0,t-10*beta)+1)
@coeff_func
def inverseSq_(t,alpha=1,beta=1):
    """beta/(t+alpha+1)**(0.5)"""
    return beta/(t+alpha+1)**(0.5)
@coeff_func
def periodic_(t, alpha=1,beta=1):
    """abs(0.5*beta*sin(0.1*alpha*t + alpha*beta))"""
    return 0.5*beta*np.sin(0.1*alpha*t)
@coeff_func
def constant_(t,alpha=1,beta=1):
    """beta"""
    return beta

FUNCS = [exp_comp_, log_, inverseSq_, constant_, periodic_ ] 


def right_censor_lifetimes(lifetimes, max_ , min_ = 0):
  """
  Right censor the deaths, uniformly
    lifetimes: (n,) array of positive random variables
    max_: the max time a censorship can occur
    min_: the min time a censorship can occur 

  Returns 
    The actual observations including uniform right censoring, and
    D_i (observed death or did not)

  I think this is deprecated
  """
  n = lifetimes.shape[0]
  u = min_ + (max_-min_)*random.rand(n)
  observations = np.minimum( u, lifetimes)
  return observations, lifetimes == observations


### Models with covariates

def generate_covariates(n, d, n_binary=0, p=0.5):
    """
    n: the number of instances, integer
    d: the dimension of the covarites, integer
    binary: a float between 0 and d the represents the binary covariates 
    p: in binary, the probability of 1

    returns (n, d+1)
    """
    assert ( n_binary>=0 and n_binary<=d ), "binary must be between 0 and d"
    covariates = np.zeros((n,d+1))
    covariates[:, :d-n_binary] = random.exponential(1, size=(n, d-n_binary))
    covariates[:, d-n_binary:-1 ] =  random.binomial(1, p, size=(n, n_binary))
    covariates[:,-1] = np.ones(n)
    return covariates



def constant_coefficients(d, timelines, constant=False, independent=0):
    """
    Proportional hazards model.

    d: the dimension of the dataset
    timelines: the observational times
    constant: True for constant coefficients 
    independent: the number of coffients to set to 0 (covariate is ind of survival), or 
      a list of covariates to make indepent.

    returns a matrix (t,d+1) of coefficients
    """
    return time_varying_coefficients(d, timelines, constant=True, independent=independent, randgen=random.normal )


def time_varying_coefficients(d, timelines, constant=False, independent=0, randgen= random.exponential):
    """
    Time vary coefficients

    d: the dimension of the dataset
    timelines: the observational times
    constant: True for constant coefficients 
    independent: the number of coffients to set to 0 (covariate is ind of survival), or 
      a list of covariates to make indepent.
    randgen: how scalar coefficients (betas) are sampled.

    returns a matrix (t,d+1) of coefficients
    """
    t = timelines.shape[0]
    try:
      a = np.arange(d)
      random.shuffle(a)
      independent = a[:independent]
    except IndexError:
      pass

    n_funcs = len(FUNCS)
    coefficients = np.zeros((t,d))
    data_generators = []
    for i in range(d):
       f = FUNCS[random.randint(0,n_funcs)] if not constant else constant_
       if i in independent:
          beta = 0
       else:
          beta = randgen((1-constant)*0.5/d)
       coefficients[:,i] = f(timelines,alpha=randgen(2000.0/t), beta=beta)
       data_generators.append(f.__doc__)

    df_coefficients = pd.DataFrame( coefficients, columns = data_generators, index = timelines)
    return df_coefficients


def generate_hazard_rates(n,d, timelines, constant=False, independent=0, n_binary=0, model="aalen"):
  """
    n: the number of instances
    d: the number of covariates
    lifelines: the observational times
    constant: make the coeffients constant (not time dependent)
    n_binary: the number of binary covariates
    model: from ["aalen", "cox"]

  Returns:s
    hazard rates: (t,n) dataframe,
    coefficients: (t,d+1) dataframe of coefficients, 
    covarites: (n,d) dataframe

  """
  covariates = generate_covariates(n,d, n_binary=n_binary)
  if model=="aalen":
      coefficients = time_varying_coefficients(d+1,timelines, independent=independent, constant=constant)
      hazard_rates = np.dot( covariates, coefficients.T )
      return pd.DataFrame(hazard_rates.T,index=timelines), coefficients, pd.DataFrame(covariates)
  elif model=="cox":
      covariates = covariates[:,:-1]
      coefficients = constant_coefficients(d, timelines, independent)
      baseline = time_varying_coefficients(1, timelines)
      hazard_rates =np.exp(np.dot( covariates, coefficients.T ))*baseline[baseline.columns[0]].values
      coefficients[ "baseline: "+ baseline.columns[0] ] = baseline.values
      return pd.DataFrame(hazard_rates.T,index=timelines),coefficients, pd.DataFrame(covariates)
  else:
    raise Exception

def generate_random_lifetimes(hazard_rates,timelines, size = 1, censor=None):
  """ 
  Based on the hazard rates, compute random variables from the survival function
    hazard_rates: (n,t) array of hazard rates
    timelines: (t,) the observation times
    size: the number to return, per hardard rate 
    censor: If True, adds uniform censoring between timelines.max() and  0
            If a postive number, censors all events above that value.
            If (n,) np.array >=0 , censor elementwise. 


  Returns:
    survival_times: (size,n) array of random variables.
    (optional) censorship: if censor is true, returns (size,n) array with bool True 
       if the death was observed (not right-censored)
  """
  n = hazard_rates.shape[1]
  survival_times = np.empty((n,size))
  cumulative_hazards = cumulative_quadrature(hazard_rates.values.T, timelines)

  for i in range(size):
     u = random.rand(n,1)
     e = -np.log(u)
     v = (e - cumulative_hazards) < 0
     cross = v.argmax(1)
     survival_times[:,i] = timelines[cross]
     survival_times[cross==0,i] = np.inf

  if censor is not None:
      if type(censor)==bool:
        T = timelines.max()
        rv = T*random.uniform(size=survival_times.shape)
      else:
        rv = censor

      observed = np.less_equal(survival_times, rv)
      survival_times = np.minimum(rv, survival_times)
      return survival_times.T, observed.T
  else:
      return survival_times

def generate_observational_matrix(n,d, timelines, constant=False, independent=0, n_binary=0, model="aalen"):
  hz, coeff, covariates = generate_hazard_rates(n,d, timelines, constant=False, independent=0, n_binary=0, model="aalen")
  R = generate_random_lifetimes(hz,timelines)
  covariates["event_at"] = R.T
  return covariates.sort("event_at"), pd.DataFrame(cumulative_quadrature(coeff.values.T,timelines).T, columns = coeff.columns, index = timelines )

def cumulative_quadrature(fx, x):
  """Boss algo"""

  lower_x = x[0]
  cum_integral = np.empty((fx.shape[0],x.shape[0]))
  cum_integral[:,0] = 0
  for i in range(1,x.shape[0]):
    if (x[i-1]-lower_x) == 0:
        cum_integral[:,i] = 0.5*(fx[:,i-1] + fx[:,i])*(x[i] - lower_x)/i
        continue
    cum_integral[:,i] = ( cum_integral[:,i-1]*( (i-1)/(x[i-1]-lower_x) ) + 0.5*fx[:,i-1] + 0.5*fx[:,i])*( (x[i] - lower_x)/i )
  return cum_integral

def construct_survival_curves(hazard_rates, timelines):
  """
  Given hazard rates, reconstruct the survival curves 
    hazard_rates: (n,t) array
    timelines: (t,) the observational times 

  Returns:t
    survial curves, (n,t) array
  """
  #pdb.set_trace()
  cumulative_hazards = cumulative_quadrature(hazard_rates.values.T, timelines).T
  return pd.DataFrame( np.exp(-cumulative_hazards), index=timelines)



